from copy import deepcopy

from pkglts.config_management import Config


def test_pkg_url_empty_default():
    cfg = Config(dict(pysetup={"intended_versions": ["2.7"]}))
    cfg.load_extra()
    assert cfg._env.globals['pysetup'].pkg_url == ""


def test_pkg_url_look_multiple_places():
    tpl_cfg = dict(base={'pkgname': 'toto', 'namespace': 'oa', 'url': None},
                   doc={'fmt': 'rst'},
                   github={'url': None, 'project': 'project', 'owner': 'toto'},
                   gitlab={'url': None, 'project': 'project', 'owner': 'toto'},
                   pypi={'classifiers': [], 'servers': [dict(name="pypi", url="https://upload.pypi.org/legacy/")]},
                   readthedocs={'project': 'project'},
                   pysetup={'intended_versions': ["2.7"]})

    for name in ("base", "github", "gitlab"):  # TODO, "pypi", "readthedocs"):
        loc_cfg = deepcopy(tpl_cfg)
        loc_cfg[name]['url'] = name
        cfg = Config(loc_cfg)
        cfg.load_extra()
        assert cfg._env.globals['pysetup'].pkg_url == name


def test_py_min_ver_uses_min_intended_version():
    cfg = Config(dict(pysetup={"intended_versions": ["2.7"]}))
    cfg.load_extra()
    assert cfg._env.globals['pysetup'].py_min_ver == "2.7"

    cfg = Config(dict(pysetup={"intended_versions": ["2.7", "3.6"]}))
    cfg.load_extra()
    assert cfg._env.globals['pysetup'].py_min_ver == "2.7"

    cfg = Config(dict(pysetup={"intended_versions": ["3.6", "3.7"]}))
    cfg.load_extra()
    assert cfg._env.globals['pysetup'].py_min_ver == "3.6"

    cfg = Config(dict(pysetup={"intended_versions": ["3.9", "3.10"]}))
    cfg.load_extra()
    assert cfg._env.globals['pysetup'].py_min_ver == "3.9"



from pkglts.config_management import Config


def test_badge():
    cfg = Config(dict(coveralls={},
                      doc={'fmt': 'rst'},
                      github={'owner': "moi", 'project': "project"}))
    cfg.load_extra()
    assert ".. image:" in cfg._env.globals['coveralls'].badge



from argparse import ArgumentParser
from pathlib import Path

import pkglts
from pkglts.config_management import Config
from pkglts.option.doc.option import OptionDoc
from pkglts.option.sphinx.option import OptionSphinx
from pkglts.option.test.option import OptionTest
from pkglts.option_tools import available_options


def pkglts_opts():
    # walk through all possible options defined by pkglts
    option_basedir = Path(pkglts.__file__).parent / 'option'
    for pth in option_basedir.glob("*/"):
        option_name = pth.name
        if pth.is_dir() and not option_name.startswith("_"):
            try:
                yield available_options[option_name]
            except KeyError:
                assert False


def test_version_is_defined():
    for opt in pkglts_opts():
        assert opt.version() != "0.0.0"


def test_root_dir_is_defined():
    for opt in pkglts_opts():
        assert opt.root_dir() is not None


def test_options_expose_parameters():
    for opt in pkglts_opts():
        cfg = {}
        try:
            opt.update_parameters(cfg)
            assert len(cfg) == 1
        except NotImplementedError:  # for options that depends on cfg
            pass


def test_require_correctly_defined():
    cfg = Config(dict(base={}))
    OptionDoc('doc').update_parameters(cfg)
    OptionTest('test').update_parameters(cfg)
    OptionSphinx('sphinx').update_parameters(cfg)
    OptionSphinx('github').update_parameters(cfg)

    for opt in pkglts_opts():
        assert len(tuple(opt.require_option(cfg))) >= 0
        assert len(tuple(opt.require(cfg))) >= 0


def test_tools_correctly_defined():
    cfg = Config(dict(base={}))

    parser = ArgumentParser(description="Package structure manager")
    subparsers = parser.add_subparsers(dest='subcmd', help="sub-command help")

    for opt in pkglts_opts():
        for tool in opt.tools(cfg):
            name, action = tool(subparsers)
            assert name.isidentifier()
            assert callable(action)



import pytest
from pkglts.config_management import Config
from pkglts.option.version.option import OptionVersion


@pytest.fixture()
def opt():
    return OptionVersion('version')


@pytest.fixture()
def cfg():
    return Config()


def test_update_parameters(opt, cfg):
    opt.update_parameters(cfg)
    assert len(cfg['version']) == 3


def test_config_check_version_numbers_are_valid(opt, cfg):
    cfg['version'] = {'major': "", 'minor': "", 'post': ""}
    assert 'version.major' in opt.check(cfg)
    assert 'version.minor' in opt.check(cfg)
    assert 'version.post' in opt.check(cfg)
    cfg['version'] = {'major': "a", 'minor': "a", 'post': "a"}
    assert 'version.major' in opt.check(cfg)
    assert 'version.minor' in opt.check(cfg)
    assert 'version.post' in opt.check(cfg)
    cfg['version'] = {'major': "1", 'minor': "1", 'post': "1"}
    assert 'version.major' in opt.check(cfg)
    assert 'version.minor' in opt.check(cfg)
    assert 'version.post' in opt.check(cfg)
    cfg['version'] = {'major': 1, 'minor': 0, 'post': "2.dev"}
    assert 'version.post' in opt.check(cfg)


def test_require_option(opt, cfg):
    assert len(tuple(opt.require_option(cfg))) == 1


def test_require(opt, cfg):
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 0



from pathlib import Path

from pkglts.option.doc import fmt_badge
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionLgtm(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def require_option(self, cfg):
        return ['github', 'pysetup']

    def environment_extensions(self, cfg):
        owner = cfg['github']['owner']
        project = cfg['github']['project']

        url = f"https://lgtm.com/projects/g/{owner}/{project}/context:python"
        img = f"https://img.shields.io/lgtm/grade/python/g/{owner}/{project}.svg?logo=lgtm&logoWidth=18"
        badge = fmt_badge(img, url, "Language grade: Python", cfg['doc']['fmt'])

        return {"badge": badge}



from pkglts.config_management import Config


def test_badge():
    cfg = Config(dict(base={'pkgname': "pkg", 'namespace': None,
                            "url": "http://toto"},
                      doc={'fmt': 'rst'},
                      pysetup={"intended_versions": ["2.7", "2.8"],
                               'require': []},
                      pypi={'classifiers': [], 'servers': [dict(name="pypi", url="https://upload.pypi.org/legacy/")]}))
    cfg.load_extra()
    assert ".. image:" in cfg._env.globals['pypi'].badge


def test_auto_classifiers():
    cfg = Config(dict(base={'pkgname': "pkg", 'namespace': None,
                            "url": "http://toto"},
                      doc={'fmt': 'rst'},
                      pysetup={"intended_versions": ["2.7", "2.8"],
                               'require': []},
                      pypi={'classifiers': [], 'servers': []}))
    cfg.load_extra()
    section = cfg._env.globals['pypi']
    assert "Programming Language :: Python" in section.auto_classifiers
    assert "Programming Language :: Python :: 2.7" in section.auto_classifiers



from pathlib import Path

import pytest
from pkglts.option.reqs.find_requirements import find_reqs, action_find_reqs
from pkglts.small_tools import ensure_created, rmdir
from pkglts.config_management import Config


@pytest.fixture()
def tmp_dir():
    pth = Path("tagadareqs")

    ensure_created(pth)

    yield pth

    rmdir(pth)


def test_find_reqs_parses_files(tmp_dir):
    pth = tmp_dir / "toto.py"
    with open(pth, 'w') as fhw:
        fhw.write("import os\n"
                  "from sys import argv\n"
                  "\n"
                  "import pytest\n"
                  "import numpy as np\n"
                  "from pandas import DataFrame\n"
                  "from toto import failure\n"
                  "\n"
                  "a = 1\n"
                  "\n"
                  "from doom import pestilence\n"
                  "")

    reqs = find_reqs(pth)

    for name in ('os', 'sys', 'pytest', 'numpy', 'pandas', 'toto', 'doom'):
        assert name in reqs


def test_myoutput(tmp_dir, capsys):
    pth = tmp_dir / "toto.py"
    with open(pth, 'w') as fhw:
        fhw.write("import os\n"
                  "from sys import argv\n"
                  "\n"
                  "import pytest\n"
                  "import numpy as np\n"
                  "from pandas import DataFrame\n"
                  "from toto import failure\n"
                  "\n"
                  "a = 1\n"
                  "\n"
                  "from doom import pestilence\n"
                  "")

    cfg = Config(dict(base={'pkgname': "walou",
                            'namespace': None,
                            'owner': 'moi',
                            'url': None}))
    action_find_reqs(cfg)

    out, err = capsys.readouterr()
    assert "standard" in out



from pathlib import Path

from pkglts.option_object import Option
from pkglts.version import __version__


class OptionPysetup(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        sec = dict(
            intended_versions=["3.10"],
        )
        cfg[self._name] = sec

    def check(self, cfg):
        invalids = []
        intended_versions = cfg[self._name]['intended_versions']

        if not intended_versions:
            invalids.append("pysetup.intended_versions")

        return invalids

    def require_option(self, cfg):
        return ['src', 'doc', 'license', 'version', 'reqs']

    def environment_extensions(self, cfg):
        py_min_ver = sorted([int(v) for v in ver.split(".")] for ver in cfg[self._name]['intended_versions'])[0]

        universal = len(set(ver.split(".")[0] for ver in cfg[self._name]['intended_versions'])) > 1

        return {
            "pkg_url": pkg_url(cfg),
            "py_min_ver": ".".join(str(v) for v in py_min_ver),
            "universal": universal
        }


def pkg_url(cfg):
    """Extract a valid url from all config sections.

    Args:
        cfg (Config):  current package configuration

    Returns:
        (str): a valid url for the package
    """
    try:
        url = cfg['base']['url']
        if url is not None:
            return url
    except KeyError:
        pass

    try:
        url = cfg['github']['url']
        if url is not None:
            return url
    except KeyError:
        pass

    try:
        url = cfg['gitlab']['url']
        if url is not None:
            return url
    except KeyError:
        pass

    # try:
    #     url = cfg['pypi']['url']
    #     if url is not None:
    #         return url
    # except KeyError:
    #     pass

    # try:
    #     url = cfg['readthedocs']['url']
    #     if url is not None:
    #         return url
    # except KeyError:
    #     pass

    return ""



from pkglts.templating import parse_source


def test_parse_finds_all_blocks():
    txt = ("assert 'toto'\n"
           "assert 'titi'\n"
           "# {# pkglts, bid\n"
           "this is content\n"
           "# #}\n"
           "after block")
    blocks = parse_source(txt)
    assert len(blocks) == 3
    b1, b2, b3 = blocks
    assert b1.before_header == ""
    assert b1.bid is None
    assert b1.after_header == ""
    assert b1.content == ("assert 'toto'\n"
                          "assert 'titi'\n")
    assert b1.before_footer == ""
    assert b1.after_footer == ""

    assert b2.before_header == "# "
    assert b2.bid == "bid"
    assert b2.after_header == ""
    assert b2.content == "this is content\n"
    assert b2.before_footer == "# "
    assert b2.after_footer == ""

    assert b3.before_header == ""
    assert b3.bid is None
    assert b3.after_header == ""
    assert b3.content == "after block"
    assert b3.before_footer == ""
    assert b3.after_footer == ""


def test_parse_finds_bef_after_header():
    txt = ("assert 'toto'\n"
           "assert 'titi'\n"
           "# bef header{# pkglts, bid after header\n"
           "this is content\n"
           "# #}\n"
           "after block")
    blocks = parse_source(txt)
    assert len(blocks) == 3
    _, block, _ = blocks
    assert block.before_header == "# bef header"
    assert block.bid == "bid"
    assert block.after_header == " after header"
    assert block.content == "this is content\n"
    assert block.before_footer == "# "
    assert block.after_footer == ""


def test_parse_finds_bef_after_footer():
    txt = ("assert 'toto'\n"
           "assert 'titi'\n"
           "# {# pkglts, bid\n"
           "this is content\n"
           "# bef footer#} after footer\n"
           "after block")
    blocks = parse_source(txt)
    assert len(blocks) == 3
    _, block, _ = blocks
    assert block.before_header == "# "
    assert block.bid == "bid"
    assert block.after_header == ""
    assert block.content == "this is content\n"
    assert block.before_footer == "# bef footer"
    assert block.after_footer == " after footer"


def test_parse_finds_single_pkglts_block():
    txt = ("# {# pkglts, bid\n"
           "this is content\n"
           "# bef footer#} after footer\n")
    blocks = parse_source(txt)
    assert len(blocks) == 1
    block, = blocks
    assert block.before_header == "# "
    assert block.bid == "bid"
    assert block.after_header == ""
    assert block.content == "this is content\n"
    assert block.before_footer == "# bef footer"
    assert block.after_footer == " after footer"



import logging
from datetime import datetime
from pathlib import Path

from pkglts.option_object import Option
from pkglts.version import __version__

LOGGER = logging.getLogger(__name__)

TPL_DIR = Path(__file__).parent / "templates"


class OptionLicense(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        sec = dict(
            name="cecill-c",
            year=datetime.now().year,
            organization="organization",
            project="{{ base.pkgname }}"
        )
        cfg[self._name] = sec

    def check(self, cfg):
        invalids = []
        name = cfg[self._name]['name']
        # year = pkg_cfg['license']['year']
        # organization = pkg_cfg['license']['organization']
        # project = pkg_cfg['license']['project']

        if not name or not get_tpl_path(name).exists():
            invalids.append('license.name')

        return invalids

    def require_option(self, cfg):
        return ['base']

    def environment_extensions(self, cfg):
        return {"full_text": full_text(cfg)}


def get_tpl_path(name):
    """Return a path for license template file.

    Warnings: Do not test if path exists

    Args:
        name (str): name of license to fetch

    Returns:
        (str)
    """
    return TPL_DIR / f"{name.lower()}.txt"


def full_text(cfg):
    """Generate a license
    """
    name = cfg['license']['name']

    # open template
    try:
        with open(get_tpl_path(name), 'r') as fhr:
            cnt = fhr.read()

        return cfg.render(cnt)
    except FileNotFoundError as err:
        LOGGER.error("unable to find template for given license")
        raise err



"""
Main config object and functions to manipulate it.
"""
import json
import logging
from datetime import date

from jinja2 import Environment, StrictUndefined, UndefinedError

from .config import pkg_cfg_file, pkglts_dir
from .option.git.option import OptionGit
from .option.pypi.option import OptionPypi
from .option.reqs.option import OptionReqs
from .option.src.option import OptionSrc
from .option_tools import available_options, find_available_options

CURRENT_PKG_CFG_VERSION = 17

LOGGER = logging.getLogger(__name__)

DEFAULT_CFG = dict(_pkglts=dict(use_prompts=False,
                                auto_install=True,
                                install_front_end='stdout',
                                version=CURRENT_PKG_CFG_VERSION))

find_available_options()


class ConfigSection(object):
    """Small class to allow accessing parameters using the dot
    method instead of ['param_name'] method
    """
    pass


class Config(dict):
    """Object used to store both a templated version of the config as a dict interface
    its resolution and a jinja2 environment that reflect the config.
    """

    def __init__(self, *args, **kwds):
        dict.__init__(self)
        self._tpl = dict(*args, **kwds)

        # initialise associated Jinja2 environment
        self._env = Environment(undefined=StrictUndefined)
        self._env.keep_trailing_newline = True

        # add global filters and test
        self._env.globals['today'] = lambda: date.today().isoformat()

        self.add_test('available', self._is_available)

        # resolve
        self.resolve()

    def __eq__(self, other):
        return dict.__eq__(self, other)

    def __ne__(self, other):
        return dict.__ne__(self, other)

    def template(self):
        """Associated template even after resolution."""
        return self._tpl

    def add_test(self, name, func):
        """Add a new test in jinja2 environment.

        Args:
            name (str): name of test (must be unique)
            func (callable): function use for test

        Returns:
            None
        """
        self._env.tests[name] = func

    def _is_available(self, opt_name):
        return opt_name in self

    def _add_param(self, opt_name, param_name, param_value):
        """Add a new parameter value in the config

        Args:
            opt_name (str): Option name
            param_name (str): parameter name
            param_value (any): parameter value

        Returns:
            None
        """
        self[opt_name][param_name] = param_value
        setattr(self._env.globals[opt_name], param_name, param_value)

    def resolve(self):
        """Try to resolve all templated items.

        Returns:
            None
        """
        to_eval = []
        for opt_name, cfg in self._tpl.items():
            self[opt_name] = {}
            self._env.globals[opt_name] = ConfigSection()
            for key, param in cfg.items():
                if isinstance(param, str):
                    to_eval.append((opt_name, key, param))
                else:
                    self._add_param(opt_name, key, param)

        nb_iter_max = len(to_eval) ** 2
        cur_iter = 0
        while to_eval and cur_iter < nb_iter_max:
            cur_iter += 1
            opt_name, key, param = to_eval.pop(0)
            try:
                txt = self.render(param)
                self._add_param(opt_name, key, txt)
            except UndefinedError:
                to_eval.append((opt_name, key, param))

        if to_eval:
            msg = "unable to fully render config\n"
            for item in to_eval:
                msg += "{}:{} '{}'\n".format(*item)
            raise UserWarning(msg)

    def load_extra(self):
        """load option specific handlers.

        Returns:
            None
        """
        for opt_name in self.installed_options():
            try:
                opt = available_options[opt_name]
                for func_name, func in opt.environment_extensions(self).items():
                    setattr(self._env.globals[opt_name], func_name, func)
            except KeyError:
                raise KeyError(f"option '{opt_name}' exists in config but does not appear to be installed")

    def installed_options(self, return_sorted=False):
        """List all installed options.

        Returns:
            (iter of str)
        """
        installed = [key for key in self if not key.startswith("_")]

        if return_sorted:
            opt_names = installed
            installed = []
            while opt_names:
                name = opt_names.pop(0)
                opt = available_options[name]
                if any(dep not in installed for dep in opt.require_option(self)):
                    opt_names.append(name)
                else:
                    installed.append(name)

        for name in installed:
            yield name

    def render(self, txt):
        """Use items in config to render text

        Args:
            txt (str): templated text to render

        Returns:
            (str): same text where all templated parts have been replaced
                   by their values.
        """
        return self._env.from_string(txt).render()


def get_pkg_config(rep="."):
    """Read pkg_cfg file associated to this package.

    Args:
        rep (Path): directory to search for info

    Returns:
        (Config): Config initialized with pkg_config
    """
    with open(rep / pkglts_dir / pkg_cfg_file, 'r') as fhr:
        pkg_cfg = json.load(fhr)

    # update version of pkg_config
    file_version = pkg_cfg['_pkglts'].get('version', 0)
    for i in range(file_version, CURRENT_PKG_CFG_VERSION):
        upgrade_pkg_cfg_version(pkg_cfg, i)

    # create Config object
    cfg = Config(pkg_cfg)
    cfg.load_extra()

    # write back config if version has been updated
    if file_version < CURRENT_PKG_CFG_VERSION:
        write_pkg_config(cfg, rep)

    return cfg


def write_pkg_config(cfg, rep="."):
    """Store config associated to this package on disk.

    Args:
        cfg (Config): current working config
        rep (Path): directory to search for info

    Returns:
        None
    """
    LOGGER.info("write package config")
    pkg_cfg = dict(cfg.template())

    with open(rep / pkglts_dir / pkg_cfg_file, 'w') as fhw:
        json.dump(pkg_cfg, fhw, sort_keys=True, indent=2)


def upgrade_pkg_cfg_version(pkg_cfg, version):
    """Upgrade the version of pkg_cfg file from version to version +1

    Args:
        pkg_cfg (dict of str, any): package configuration
        version (int): current version of file

    Returns:
        (dict of str: any): a reference to an updated pkg_cfg
    """
    if version == 0:
        pkg_cfg['_pkglts']['version'] = 1
    elif version == 1:
        pkg_cfg['_pkglts']['version'] = 2
    elif version == 2:
        pkg_cfg['_pkglts']['version'] = 3
        if 'pysetup' in pkg_cfg:
            section = pkg_cfg['pysetup']
            section['require'] = section.get('require', [])
    elif version == 3:
        pkg_cfg['_pkglts']['version'] = 4
        if 'test' in pkg_cfg:
            section = pkg_cfg['test']
            section['suite_name'] = section.get('suite_name', "pytest")
    elif version == 4:
        pkg_cfg['_pkglts']['version'] = 5
        if 'base' in pkg_cfg:
            section = pkg_cfg['base']
            section['namespace_method'] = section.get('namespace_method', "pkg_util")
    elif version == 5:
        pkg_cfg['_pkglts']['version'] = 6
        if 'pysetup' in pkg_cfg:
            section = pkg_cfg['pysetup']
            deps = []
            for pkg_mng, name in section.get('require', []):
                if pkg_mng == 'none':
                    deps.append(dict(name=name))
                else:
                    deps.append(dict(name=name, pkg_mng=pkg_mng))
            section['require'] = deps
    elif version == 6:
        pkg_cfg['_pkglts']['version'] = 7
        if 'sphinx' in pkg_cfg:
            section = pkg_cfg['sphinx']
            section['build_dir'] = section.get('build_dir', "build/sphinx")
    elif version == 7:
        pkg_cfg['_pkglts']['version'] = 8
        if 'doc' in pkg_cfg:
            section = pkg_cfg['doc']
            section['fmt'] = section.get('fmt', 'rst')
    elif version == 8:
        pkg_cfg['_pkglts']['version'] = 9
        if 'github' in pkg_cfg or 'gitlab' in pkg_cfg:
            LOGGER.info("\n"
                        "################################################\n"
                        "\n"
                        "\n"
                        "please remove '.gitignore file and regenerate package\n"
                        "\n"
                        "\n"
                        "################################################\n")
            OptionGit('git').update_parameters(pkg_cfg)
    elif version == 9:
        pkg_cfg['_pkglts']['version'] = 10
        if 'pypi' in pkg_cfg:
            LOGGER.info("\n"
                        "################################################\n"
                        "\n"
                        "\n"
                        "please remove '.pypirc file and regenerate package\n"
                        "\n"
                        "\n"
                        "################################################\n")
            # save section
            mem = dict(pkg_cfg["pypi"])
            # get newly defined list of servers
            OptionPypi('pypi').update_parameters(pkg_cfg)
            servers = dict(pkg_cfg["pypi"]["servers"])
            # update option
            pkg_cfg["pypi"] = mem
            pkg_cfg["pypi"]["servers"] = pkg_cfg["pypi"].get("servers", {})
            pkg_cfg["pypi"]["servers"].update(servers)
    elif version == 10:
        pkg_cfg['_pkglts']['version'] = 11
        if 'data' in pkg_cfg:
            section = pkg_cfg['data']
            section['filetype'] = section.get('filetype', [".json", ".ini"])
            section['use_ext_dir'] = section.get('use_ext_dir', False)
    elif version == 11:
        pkg_cfg['_pkglts']['version'] = 12
        if 'pypi' in pkg_cfg:
            # save servers
            servers = dict(pkg_cfg["pypi"]["servers"])
            # transform dict into list
            servers_list = [dict(name=name, url=descr['url']) for name, descr in servers.items()]
            pkg_cfg["pypi"]["servers"] = servers_list
    elif version == 12:
        pkg_cfg['_pkglts']['version'] = 13
        if 'base' in pkg_cfg:
            namespace_method = pkg_cfg['base'].pop('namespace_method')

            if 'src' not in pkg_cfg:
                OptionSrc('src').update_parameters(pkg_cfg)

            pkg_cfg['src']['namespace_method'] = namespace_method

        if 'pysetup' in pkg_cfg:
            require = pkg_cfg['pysetup'].pop('require')

            if 'reqs' not in pkg_cfg:
                OptionReqs('reqs').update_parameters(pkg_cfg)

            pkg_cfg['reqs']['require'] = require

        if 'sphinx' in pkg_cfg:
            pkg_cfg['sphinx']['gallery'] = pkg_cfg['sphinx'].get('gallery', "")
    elif version == 13:
        pkg_cfg['_pkglts']['version'] = 14

        if 'sphinx' in pkg_cfg:
            pkg_cfg['sphinx']['doc_dir'] = pkg_cfg['sphinx'].get('doc_dir', "doc")
    elif version == 14:
        pkg_cfg['_pkglts']['version'] = 15

        if 'git' in pkg_cfg:
            try:
                perm_branches = pkg_cfg['git']['permanent_branches']
            except KeyError:
                if 'gitlab' in pkg_cfg or 'github' in pkg_cfg:
                    perm_branches = ["main"]
                else:
                    perm_branches = []

            pkg_cfg['git']['permanent_branches'] = perm_branches
    elif version == 15:
        pkg_cfg['_pkglts']['version'] = 16

        if 'conda' in pkg_cfg:
            try:
                env_name = pkg_cfg['conda']['env_name']
            except KeyError:
                env_name = "{{ base.pkgname }}"

            pkg_cfg['conda']['env_name'] = env_name
    elif version == 16:
        pkg_cfg['_pkglts']['version'] = 17

        if 'pysetup' in pkg_cfg:
            intended_versions = []
            for ver in pkg_cfg['pysetup']['intended_versions']:
                if "." in ver:
                    intended_versions.append(ver)
                else:
                    intended_versions.append(".".join((ver[0], ver[1:])))

            pkg_cfg['pysetup']['intended_versions'] = intended_versions

    return pkg_cfg



from pathlib import Path

import pytest
from pkglts.config_management import Config
from pkglts.option.notebook.option import OptionNotebook
from pkglts.small_tools import ensure_created, rmdir


@pytest.fixture()
def opt():
    return OptionNotebook('notebook')


@pytest.fixture()
def cfg():
    return Config()


@pytest.fixture()
def tmp_dir():
    pth = Path("nb_test_config")
    ensure_created(pth)

    yield pth

    rmdir(pth)


def test_update_parameters(opt, cfg):
    opt.update_parameters(cfg)
    assert len(cfg['notebook']) == 1


def test_config_check_src_directory(opt, cfg, tmp_dir):
    cfg['notebook'] = {'src_directory': "failed_nb"}
    assert 'notebook.src_directory' in opt.check(cfg)

    cfg['notebook'] = {'src_directory': tmp_dir}
    assert 'notebook.src_directory' not in opt.check(cfg)


def test_require_option(opt, cfg):
    assert len(tuple(opt.require_option(cfg))) == 1


def test_require(opt, cfg):
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 0



# {# pkglts, sphinx

"""
Configuration for sphinx
see: https://www.sphinx-doc.org/en/master/usage/configuration.html

"""
import os
import re
import sys



# Get the project root dir, which is the parent dir of this
cwd = os.getcwd()
project_root = os.path.dirname(cwd)
src_dir = os.path.abspath(os.path.join(project_root, "src"))

# Insert the project root dir as the first element in the PYTHONPATH.
# This lets us ensure that the source package is imported, and that its
# version is used.
sys.path.insert(0, os.path.join(project_root, 'src'))


# -- General configuration ---------------------------------------------

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.graphviz',
    'sphinx.ext.ifconfig',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
]

# try to add more extensions which are not default
# but still useful
# based on the fact that the extension is installed on the system

try:
    import matplotlib.sphinxext.plot_directive
    extensions.append('matplotlib.sphinxext.plot_directive')
except ImportError:
    pass



# default settings that can be redefined outside of the pkglts block
todo_include_todos = True
autosummary_generate = True
intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}
inheritance_node_attrs = dict(shape='ellipse', fontsize=12,
                              color='orange', style='filled')

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u"pkglts"
copyright = u"2015, pkglts"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#

# The short X.Y version.
version = "6.5.0"
# The full version, including alpha/beta/rc tags.
release = "6.5.0"


exclude_patterns = ['build', 'dist']

pygments_style = 'sphinx'

# -- Options for HTML output -------------------------------------------

html_theme = "sphinx_rtd_theme"
html_static_path = ['_static']
htmlhelp_basename = "pkgltsdoc"


# -- Options for LaTeX output ------------------------------------------

latex_elements = {
    # The paper size ('letterpaper' or 'a4paper').
    # 'papersize': 'letterpaper',

    # The font size ('10pt', '11pt' or '12pt').
    # 'pointsize': '10pt',

    # Additional stuff for the LaTeX preamble.
    # 'preamble': '',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    (u"index", u"pkglts.tex",
     u"pkglts Documentation",
     u"revesansparole",
     u"manual"),
]

# -- Options for manual page output ------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (u"index", u"pkglts",
     u"pkglts Documentation",
     [u"revesansparole"], 1)
]

# -- Options for Texinfo output ----------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (u"index", u"pkglts",
     u"pkglts Documentation",
     u"revesansparole",
     u"pkglts",
     u"Building packages with long term support",
     u"Miscellaneous"),
]


# use apidoc to generate developer doc
try:
    from sphinx.ext.apidoc import main
except ImportError:
    from sphinx.apidoc import main


destdir = os.path.abspath(os.path.join(project_root, "doc", "_dvlpt"))

if not os.path.isdir(destdir):
    os.makedirs(destdir)

main(['-e', '-o', destdir, '-d', '4', '-s', source_suffix[1:], '--force', src_dir])

# #}



from pkglts.option_tools import get_user_permission

loc_input = 'builtins.input'


def test_user_permission(mocker):
    mocker.patch(loc_input, return_value='')
    assert get_user_permission('action')

    mocker.patch(loc_input, return_value='y')
    assert get_user_permission('action')

    mocker.patch(loc_input, return_value='n')
    assert not get_user_permission('action')

    mocker.patch(loc_input, return_value='N')
    assert not get_user_permission('action')

# def test_get_key():
#     assert get_key('toto', {'toto': 'titi'}) == 'titi'
#
#
# def test_get_key_nested():
#     assert get_key('toto.titi', {'toto': {'titi': 'tata'}}) == 'tata'
#
#
# def test_get_key_returns_none_for_unknown_key():
#     assert get_key('tata', {'toto': {'titi': 'tata'}}) is None
#     assert get_key('toto.tata', {'toto': {'titi': 'tata'}}) is None
#     assert get_key('titi', {'toto': {'titi': 'tata'}}) is None
#
#
# def test_ask_arg_do_not_prompt_user_if_value_in_extra(mocker):
#     mocker.patch(loc_input, return_value='useless')
#     assert ask_arg('toto', None, None, {'toto': 1}) == 1
#
#
# def test_ask_arg_find_default_in_pkg_cfg(mocker):
#     mocker.patch(loc_input, return_value='')
#     assert ask_arg('toto', {'toto': 1}) == '1'
#     assert ask_arg('toto', {'titi': 1}) == ''
#     assert ask_arg('toto') == ''
#
#
# def test_ask_use_default_if_everything_fail_only(mocker):
#     mocker.patch(loc_input, return_value='')
#     assert ask_arg('toto', {'toto': 1}, 2) == '1'
#     assert ask_arg('toto', {'titi': 1}, 2) == '2'
#
#
# def test_ask_arg_user_bypass_default(mocker):
#     mocker.patch(loc_input, return_value='myvalue')
#     assert ask_arg('toto', {'toto': 1}, 0, {}) == 'myvalue'
#     assert ask_arg('toto', {}, 0, {}) == 'myvalue'
#     assert ask_arg('toto') == 'myvalue'
#
#
# def test_ask_arg_handle_list(mocker):
#     mocker.patch(loc_input, return_value='')
#     res = ask_arg('keys', {}, ["1"], {})
#     assert res == ["1"]
#     res = ask_arg('keys', {}, ["1", "2"], {})
#     assert res == ["1", "2"]
#     res = ask_arg('keys', {'keys': ["1", "2"]}, [], {})
#     assert res == ["1", "2"]
#     res = ask_arg('keys', {'keys': ["1", "2"]}, ["1"], {})
#     assert res == ["1", "2"]
#     res = ask_arg('keys', {'keys': ["1"]}, ["1", "2"], {})
#     assert res == ["1"]
#     assert_raises(UserWarning, lambda: ask_arg('keys', {'keys': "1"},
#                                                ["1", "2"], {}))



import pytest
from pkglts.config_management import Config
from pkglts.option.git.option import OptionGit


@pytest.fixture()
def opt():
    return OptionGit('git')


@pytest.fixture()
def cfg():
    return Config()


def test_update_parameters(opt, cfg):
    opt.update_parameters(cfg)
    assert len(cfg['git']) == 1


def test_config_check_nothing(opt, cfg):
    cfg['git'] = {}
    assert len(opt.check(cfg)) == 0


def test_require_option(opt, cfg):
    assert len(tuple(opt.require_option(cfg))) == 1


def test_require(opt, cfg):
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 0



# {# pkglts, pysetup.kwds
# format setup arguments
from pathlib import Path
from setuptools import setup, find_packages

short_descr = "Building packages with long term support"
readme = open('README.rst').read()
history = open('HISTORY.rst').read()

# find packages
pkgs = find_packages('src')

src_dir = Path("src/pkglts")

data_files = []
for pth in src_dir.rglob("*"):
    if not pth.is_dir() and "__pycache__" not in pth.parts:
        if pth.suffix in ['', '.bat', '.cfg', '.json', '.md', '.in', '.ini', '.png', '.ps1', '.rst', '.sh', '.tpl', '.txt', '.yaml', '.yml']:
            data_files.append(str(pth.relative_to(src_dir)))

pkg_data = {'pkglts': data_files}

setup_kwds = dict(
    name='pkglts',
    version="6.5.0",
    description=short_descr,
    long_description=readme + '\n\n' + history,
    author="revesansparole",
    author_email="revesansparole@gmail.com",
    url='https://github.com/revesansparole/pkglts',
    license='CeCILL-C',
    zip_safe=False,

    packages=pkgs,
    
    package_dir={'': 'src'},
    
    
    package_data=pkg_data,
    setup_requires=[
        "pytest-runner",
        ],
    install_requires=[
        "jinja2",
        "requests",
        "semver",
        "unidecode",
        ],
    tests_require=[
        "coverage",
        "coveralls",
        "pytest",
        "pytest-cov",
        "pytest-mock",
        "tox",
        ],
    entry_points={},
    keywords='packaging, package builder',
    
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    )
# #}
# change setup_kwds below before the next pkglts tag

setup_kwds['entry_points']['console_scripts'] = ['pmg = pkglts.manage_script:main']
setup_kwds['entry_points']['pkglts'] = [
    'base = pkglts.option.base.option:OptionBase',
    'appveyor = pkglts.option.appveyor.option:OptionAppveyor',
    'conda = pkglts.option.conda.option:OptionConda',
    'coverage = pkglts.option.coverage.option:OptionCoverage',
    'coveralls = pkglts.option.coveralls.option:OptionCoveralls',
    'data = pkglts.option.data.option:OptionData',
    'doc = pkglts.option.doc.option:OptionDoc',
    'flake8 = pkglts.option.flake8.option:OptionFlake8',
    'git = pkglts.option.git.option:OptionGit',
    'github = pkglts.option.github.option:OptionGithub',
    'gitlab = pkglts.option.gitlab.option:OptionGitlab',
    'landscape = pkglts.option.landscape.option:OptionLandscape',
    'lgtm = pkglts.option.lgtm.option:OptionLgtm',
    'license = pkglts.option.license.option:OptionLicense',
    'notebook = pkglts.option.notebook.option:OptionNotebook',
    'plugin_project = pkglts.option.plugin_project.option:OptionPluginProject',
    'pypi = pkglts.option.pypi.option:OptionPypi',
    'pysetup = pkglts.option.pysetup.option:OptionPysetup',
    'readthedocs = pkglts.option.readthedocs.option:OptionReadthedocs',
    'reqs = pkglts.option.reqs.option:OptionReqs',
    'requires = pkglts.option.requires.option:OptionRequires',
    'sphinx = pkglts.option.sphinx.option:OptionSphinx',
    'src = pkglts.option.src.option:OptionSrc',
    'test = pkglts.option.test.option:OptionTest',
    'tox = pkglts.option.tox.option:OptionTox',
    'travis = pkglts.option.travis.option:OptionTravis',
    'version = pkglts.option.version.option:OptionVersion',
]

# do not change things below
# {# pkglts, pysetup.call
setup(**setup_kwds)
# #}



import pytest
from pkglts.config_management import Config
from pkglts.option.conda.option import OptionConda


@pytest.fixture()
def opt():
    return OptionConda('conda')


@pytest.fixture()
def cfg():
    return Config()


def test_update_parameters(opt, cfg):
    opt.update_parameters(cfg)
    assert len(cfg['conda']) == 1


def test_config_check_env_name_exists(opt, cfg):
    cfg['conda'] = {'env_name': "myname"}
    assert cfg['conda']['env_name'] == "myname"
    assert 'conda.env_name' not in opt.check(cfg)


def test_config_check_env_name_valid(opt, cfg):
    cfg['conda'] = {'env_name': ""}
    assert 'conda.env_name' in opt.check(cfg)

    cfg['conda'] = {'env_name': "\ttoto "}
    assert 'conda.env_name' in opt.check(cfg)

    cfg['conda'] = {'env_name': "toto-tutu"}
    assert 'conda.env_name' in opt.check(cfg)


def test_require_option(opt, cfg):
    assert len(tuple(opt.require_option(cfg))) == 1


def test_require(opt, cfg):
    opt.update_parameters(cfg)

    assert len(tuple(opt.require(cfg))) == 0



"""
Set of function to associate hash to blobs.
"""
import json
import logging
from base64 import b64encode
from hashlib import sha512
from pathlib import Path

from .config import pkg_hash_file, pkglts_dir
from .templating import parse_source

LOGGER = logging.getLogger(__name__)


def pth_as_key(pth):
    """Normalize path to enable to use them as keys

    Args:
        pth (Path): path to normalize

    Returns:
        (str)
    """
    pth = pth.resolve()
    if pth.is_absolute():  # make path relative to project root
        pth = pth.relative_to(Path.cwd())
    return pth.as_posix()


def compute_hash(txt):
    """Compute hash summary of a text

    Args:
        txt (str): content to hash

    Returns:
        (str): hash key
    """
    algo = sha512()
    algo.update(txt.encode('utf-8'))  # TODO bad if non utf-8 encoded file
    return b64encode(algo.digest()).decode('utf-8')


def get_pkg_hash(rep="."):
    """Read pkg_hash file associated to this package

    Args:
        rep (Path): directory to search for info

    Returns:
        (dict of str, list): hash map of preserved sections in this
                             package
    """
    return json.load(open(rep / pkglts_dir / pkg_hash_file, 'r'))


def write_pkg_hash(pkg_hash, rep="."):
    """Store hash associated to this package on disk.

    Args:
        pkg_hash (dict of str, list): hash map of preserved sections in this
                                     package
        rep (Path): directory to search for info

    Returns:
        None
    """
    LOGGER.info("write package hash")
    cfg = {pth_key: dict(sec_hash) for pth_key, sec_hash in pkg_hash.items()}

    with open(rep / pkglts_dir / pkg_hash_file, 'w') as fhw:
        json.dump(cfg, fhw, sort_keys=True, indent=2)


def modified_file_hash(pth, pkg_hash):
    """Check whether a file complies with previously stored hash

    Args:
        pth (Path): path to file to test
        pkg_hash (dict of str, list): hash map of preserved sections in this
                                     package

    Returns:
        (bool): whether this file has been modified or not
    """
    key = pth_as_key(pth)
    if key not in pkg_hash:
        raise IOError(f"{key} not under check")

    ref_blocks = pkg_hash[key]

    blocks = parse_source(pth.read_text())

    lts_blocks = dict((block.bid, block.content) for block in blocks
                      if block.bid is not None)
    if set(lts_blocks) != set(ref_blocks):
        return True

    for bid, cnt in lts_blocks.items():
        sha = compute_hash(cnt)
        if sha != ref_blocks[bid]:
            return True

    return False



from pathlib import Path

from pkglts.local import pkg_full_name
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionBase(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def update_parameters(self, cfg):
        sec = dict(
            pkgname=Path.cwd().name,
            namespace=None,
            url=None,
            authors=[("moi", "moi@email.com")]
        )
        cfg[self._name] = sec

    def check(self, cfg):
        invalids = []
        pkgname = cfg[self._name]['pkgname']
        namespace = cfg[self._name]['namespace']

        if "." in pkgname:
            invalids.append('base.pkgname')
        elif not pkgname.isidentifier():
            invalids.append('base.pkgname')

        if namespace is not None:
            if "." in namespace:
                invalids.append('base.namespace')
            elif not namespace.isidentifier():
                invalids.append('base.namespace')

        return invalids

    def environment_extensions(self, cfg):
        return {"pkg_full_name": pkg_full_name(cfg)}



from pathlib import Path

from pkglts.dependency import Dependency
from pkglts.option_object import Option
from pkglts.version import __version__


class OptionFlake8(Option):
    def version(self):
        return __version__

    def root_dir(self):
        return Path(__file__).parent

    def require_option(self, cfg):
        return ['base']

    def require(self, cfg):
        del cfg
        yield Dependency('flake8', intent='dvlpt')
